/* Puzzle Solver 1.0 Viktor Engelmann */

#include<iostream>
#include<vector>
#include<iomanip>
#include<fstream>
using namespace std;

class Puzzlepiece {
	public:
		int up;
		int down;
		int left;
		int right;
		Puzzlepiece(int u=0, int r=0, int d=0, int l=0) {up=u;left=l;down=d;right=r;}
		void Rotate() {*this = Puzzlepiece(left,up,right,down);}
		bool FitsRightOf(const Puzzlepiece& x) const {return((x.right == -left)||(x.right == 0));}
		bool FitsUnder(const Puzzlepiece& x)   const {return((x.down == -up)||(x.down == 0));}
};
istream& operator>>(istream& is, Puzzlepiece& P) {
	return is >> P.up >> P.right >> P.down >> P.left;
}

int w=3, h=3, n=9;
vector<vector<Puzzlepiece> > Desk;
vector<Puzzlepiece> Pieces;
vector<bool> Availiable;
int recursions=0;

void Print() {
	for(int y = 1; y <= h; y++) {
		for(int x = 1; x <= w; x++)
			cout << "  +--------+ ";
		cout << '\n';
		for(int x = 1; x <= w; x++)
			cout << "  |   " << setw(2) << Desk[y][x].up << "   | ";
		cout << '\n';
		for(int x = 1; x <= w; x++)
			cout << "  | " << setw(2) << Desk[y][x].left << "  " << setw(2) << Desk[y][x].right << " | ";
		cout << '\n';
		for(int x = 1; x <= w; x++)
			cout << "  |   " << setw(2) << Desk[y][x].down << "   | ";
		cout << '\n';
		for(int x = 1; x <= w; x++)
			cout << "  +--------+ ";
		cout << "\n\n";
	}
	cout << "\n\n"; 
}

void Puzzle(int x, int y) {
	recursions++;
	for(int i = 0; i < n; i++) {
		if(!Availiable[i]) // select unused Puzzlepiece
			continue;
		Availiable[i] = false;
		Desk[y][x] = Pieces[i];
		for(int j = 0; j < 4; j++) { // there are dummy elements on the Left and Top margin
			if(Desk[y][x].FitsRightOf(Desk[y][x-1])
			&& Desk[y][x].FitsUnder(Desk[y-1][x])) {
				if(x < w)
					Puzzle(x+1, y); // next column
				else {
					if(y < h)
						Puzzle(1, y+1); // next row
					else
						Print(); // success
				}
			}
			Desk[y][x].Rotate();
		}
		Availiable[i] = true;
	}
}

int main(int argc, char** argv) {
	if(argc < 2) {
		cerr << "usage: " << argv[0] << " <puzzle-file>\n";
		return(1);
	}

	ifstream fin(argv[1]); // get data from file
	fin >> h >> w;
	n = w*h;
	Pieces = vector<Puzzlepiece>(n);
	Availiable = vector<bool>(n,true);
	for(int i = 0; i < n; i++)
		fin >> Pieces[i];
	Desk = vector<vector<Puzzlepiece> >(h+1, vector<Puzzlepiece>(w+1));

	Puzzle(1,1); // start backtracking
	cout << "needed " << recursions << " recursions\n";
	return(0);
}





